/***
* MIT License
*
* Copyright (c) 2016 Morten Bojsen-Hansen
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
***/

#include "SIM_GasInflow.h"

#include <PRM/PRM_Include.h>
#include <SIM/SIM_DopDescription.h>
#include <SIM/SIM_PRMShared.h>
#include <SIM/SIM_RawField.h>
#include <SIM/SIM_ScalarField.h>
#include <SIM/SIM_VectorField.h>
#include <SIM/SIM_Utils.h>
#include <UT/UT_VoxelArray.h>

// BaseClass created by DECLARE_DATAFACTORY
SIM_GasInflow::SIM_GasInflow(const SIM_DataFactory* factory)
	: BaseClass(factory)
{
}

SIM_GasInflow::~SIM_GasInflow()
{
}

// set up DOP (including paramaters etc.)
const SIM_DopDescription* SIM_GasInflow::getDopDescription()
{
	static PRM_Name destinationFieldName(GAS_NAME_FIELDDEST, "Destination Field");
	static PRM_Name sourceFieldName(GAS_NAME_FIELDSOURCE, "Source Field");
	static PRM_Name velocityFieldName(GAS_NAME_VELOCITY, "Velocity Field");
	static PRM_Name layerThicknessName(PML_LAYER_THICKNESS, "Layer Thickness");
	static PRM_Name timestepMultiplierName(PML_TIMESTEP_MULTIPLIER, "Timestep Multiplier");
	static PRM_Name dampingName(PML_DAMPING, "Damping");
	static PRM_Name dampPlusXName(PML_DAMP_PLUS_X, "+X");
	static PRM_Name dampMinusXName(PML_DAMP_MINUS_X, "-X");
	static PRM_Name dampPlusYName(PML_DAMP_PLUS_Y, "+Y");
	static PRM_Name dampMinusYName(PML_DAMP_MINUS_Y, "-Y");
	static PRM_Name dampPlusZName(PML_DAMP_PLUS_Z, "+Z");
	static PRM_Name dampMinusZName(PML_DAMP_MINUS_Z, "-Z");

	static PRM_Default layerThicknessDefault(6);
	static PRM_Default timestepMultiplierDefault(1.0);

	static PRM_Template theTemplates[] = {
		PRM_Template(PRM_STRING, 1, &destinationFieldName),
		PRM_Template(PRM_STRING, 1, &sourceFieldName),
		PRM_Template(PRM_STRING, 1, &velocityFieldName),
		PRM_Template(PRM_INT, 1, &layerThicknessName, &layerThicknessDefault),
		PRM_Template(PRM_FLT, 1, &timestepMultiplierName, &timestepMultiplierDefault),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampingName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampPlusXName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampMinusXName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampPlusYName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampMinusYName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampPlusZName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE, 1, &dampMinusZName, PRMzeroDefaults),
		PRM_Template()
	};

	static SIM_DopDescription theDopDescription(
		true, // make a DOP
		"sim_gasinflow", // internal name
		"Gas Inflow", // DOP name
		"$OS", // Default data name
		classname(), // the type of this DOP, usually the class.
		theTemplates); // template list for generating the DOP

	theDopDescription.setPureApplyData(true); // no gray inputs

	return &theDopDescription;
}

bool SIM_GasInflow::solveGasSubclass(SIM_Engine &engine, SIM_Object *obj, SIM_Time time, SIM_Time timestep)
{
    UT_Interrupt *boss = UTgetInterrupt();

	SIM_VectorField *vel = getVectorField(obj, GAS_NAME_VELOCITY);
	// SIM_ScalarField *surface = getScalarField(obj, GAS_NAME_SURFACE);
	// SIM_VectorField *inflowVelocity = getVectorField(obj, GAS_NAME_INFLOW_VELOCITY);
	// SIM_ScalarField *inflowSurface = getScalarField(obj, GAS_NAME_INFLOW_SURFACE);

	SIM_DataArray destinationFields, sourceFields;
	getMatchingData(destinationFields, obj, GAS_NAME_FIELDDEST);
	getMatchingData(sourceFields, obj, GAS_NAME_FIELDSOURCE);

	if (destinationFields.entries() != sourceFields.entries())
	{
		std::cout << "#input and destination fields do not match (" << destinationFields.entries() << " vs. " << sourceFields.entries() << ")" << std::endl;
		return false;
	}

    for (int i=0; i<destinationFields.entries(); ++i)
	{
		if (boss->opInterrupt()) break;

		if (SIM_ScalarField *dst = SIM_DATA_CAST(destinationFields(i), SIM_ScalarField))
		if (SIM_ScalarField* src = SIM_DATA_CAST(sourceFields(i), SIM_ScalarField))
		{
			if (vel) inflow(dst->getField(), src->getField(), vel, timestep);
			else inflowScalar(dst->getField(), src->getField());
			dst->pubHandleModification();
		}

		if (SIM_VectorField *dst = SIM_DATA_CAST(destinationFields(i), SIM_VectorField))
		if (SIM_VectorField *src = SIM_DATA_CAST(sourceFields(i), SIM_VectorField))
		{
			if (vel)
			{
				// make temporary copy of vel to avoid writing to and reading from the same field
				SIM_VectorField *veltmp = vel;
				if (dst == vel)
				{
					veltmp = SIM_DATA_CREATE(*this, "veltmp", SIM_VectorField, SIM_DATA_RETURN_EXISTING);
					veltmp->matchField(vel, true);
					for (int dim : {0,1,2}) *veltmp->getField(dim) = *vel->getField(dim);
				}

				for (int dim : {0,1,2}) inflow(dst->getField(dim), src->getField(dim), veltmp, timestep);

				if (dst == vel)
				{
					removeNamedSubData("veltmp");
				}
			}
			else
			{
				for (int dim : {0,1,2}) inflowVector(dst->getField(dim), src->getField(dim), dim);
			}

			// flag as dirty
			dst->pubHandleModification();
		}

		if (SIM_MatrixField *dst = SIM_DATA_CAST(destinationFields(i), SIM_MatrixField))
		if (SIM_MatrixField *src = SIM_DATA_CAST(sourceFields(i), SIM_MatrixField))
		{
			// nothing to do
		}
	}

#if 0
	if (splitVelocityFields.entries() == 3 && splitInflowVelocityFields.entries() == 3)
	{
		for (int q : {0,1,2})
		{
			SIM_VectorField* dst = SIM_DATA_CAST(splitVelocityFields(q), SIM_VectorField);
			SIM_VectorField* src = SIM_DATA_CAST(splitInflowVelocityFields(q), SIM_VectorField);

			// nothing to do
			if (dst == nullptr || src == nullptr) continue;

			for (int dim : {0,1,2}) inflowVector(dst->getField(dim), src->getField(dim), dim);

			dst->pubHandleModification();
			success = true;
		}
	}
	else if (splitVelocityFields.entries() == 1 && splitInflowVelocityFields.entries() == 1)
	{
		SIM_VectorField* dst = SIM_DATA_CAST(splitVelocityFields(0), SIM_VectorField);
		SIM_VectorField* src = SIM_DATA_CAST(splitInflowVelocityFields(0), SIM_VectorField);

		if (dst != nullptr && src != nullptr)
		for (int dim : {0,1,2})
		{
			inflowVector(dst->getField(dim), src->getField(dim), dim);
		}

		dst->pubHandleModification();
		success = true;
	}
	// TODO splitVelocity==3 and inflowVelocity==1

	if (surface && inflowSurface)
	{
		inflowScalar(surface->getField(), inflowSurface->getField());
		surface->pubHandleModification();
		success = true;
	}

#endif
	return true;
}

void SIM_GasInflow::inflowVectorPartial(SIM_RawField *dst, const SIM_RawField *src, int dim, const UT_JobInfo &info)
{
    UT_Interrupt *boss = UTgetInterrupt();

	UT_VoxelArrayIteratorF vit;
	vit.setArray(dst->fieldNC());
    // vit.setCompressOnExit(true);
	vit.setPartialRange(info.job(), info.numJobs());

	UT_Vector3T<bool> dampPlus(getDampPlusX(), getDampPlusY(), getDampPlusZ());
	UT_Vector3T<bool> dampMinus(getDampMinusX(), getDampMinusY(), getDampMinusZ());

	for (vit.rewind(); !vit.atEnd(); vit.advance())
	{
	    if (vit.isStartOfTile() && boss->opInterrupt()) break;

		UT_Vector3 P;
		dst->indexToPos(vit.x(), vit.y(), vit.z(), P);

		for (int q : {0,1,2})
		{
			int L = getLayerThickness();
			int N = vit.getArray()->getRes(q);

			if (dim==q) L += 1; // staggered

			if ((dampMinus[q] && vit.idx(q) < L) || (dampPlus[q] && vit.idx(q) >= N-L))
			{
				vit.setValue(src->getValue(P));
			}
		}
	}
}

void SIM_GasInflow::inflowScalarPartial(SIM_RawField *dst, const SIM_RawField *src, const UT_JobInfo &info)
{
    UT_Interrupt *boss = UTgetInterrupt();

	UT_VoxelArrayIteratorF vit;
	vit.setArray(dst->fieldNC());
    // vit.setCompressOnExit(true);
	vit.setPartialRange(info.job(), info.numJobs());

	UT_Vector3T<bool> dampPlus(getDampPlusX(), getDampPlusY(), getDampPlusZ());
	UT_Vector3T<bool> dampMinus(getDampMinusX(), getDampMinusY(), getDampMinusZ());

	int L = getLayerThickness();

	for (vit.rewind(); !vit.atEnd(); vit.advance())
	{
	    if (vit.isStartOfTile() && boss->opInterrupt()) break;

		UT_Vector3 P;
		dst->indexToPos(vit.x(), vit.y(), vit.z(), P);

		for (int dim : {0,1,2})
		{
			int N = vit.getArray()->getRes(dim);
			if ((dampMinus[dim] && vit.idx(dim) < L) || (dampPlus[dim] && vit.idx(dim) >= N-L))
			{
				vit.setValue(src->getValue(P));
			}
		}
	}
}

void SIM_GasInflow::inflowPartial(SIM_RawField *dst, const SIM_RawField *src, const SIM_VectorField *vel, SIM_Time timestep, const UT_JobInfo &info)
{
    UT_Interrupt *boss = UTgetInterrupt();

	UT_VoxelArrayIteratorF vit;
	vit.setArray(dst->fieldNC());
    // vit.setCompressOnExit(true);
	vit.setPartialRange(info.job(), info.numJobs());

	int L = getLayerThickness();
	double multiplier = getTimestepMultiplier();

	for (vit.rewind(); !vit.atEnd(); vit.advance())
	{
	    if (vit.isStartOfTile() && boss->opInterrupt()) break;

#if 0
		UT_Vector3 P, Pold;
		UT_Vector3i idx, res;
		dst->indexToPos(vit.x(), vit.y(), vit.z(), P);
		dst->getVoxelRes(res[0], res[1], res[2]);
		res -= UT_Vector3i(1,1,1);
		Pold = P - timestep*vel->getValue(P);
		dst->posToIndex(Pold, idx[0], idx[1], idx[2]);

		// trace outside grid => inflow
		if (idx[0]-L < 0 || idx[0]+L > res[0]
		|| idx[1]-L < 0 || idx[1]+L > res[1]
		|| idx[2]-L < 0 || idx[2]+L> res[2])
		{
			vit.setValue(src->getValue(P));

			// TODO perhaps restrict to narrowband instead of allowing inflow anywhere
		}

#else
		UT_Vector3i res;
		dst->getVoxelRes(res[0], res[1], res[2]);
		res -= UT_Vector3i(1,1,1);

		// sign
		// dot(N,V) < 0 => stooooop

		// for (int sign : {-1,1})
		for (int dim : {0,1,2})
		{
			UT_Vector3 P;
			dst->indexToPos(vit.x(), vit.y(), vit.z(), P);
			UT_Vector3 N(dim==0,dim==1,dim==2);
			UT_Vector3 V(vel->getValue(P));
			fpreal val(src->getValue(P));

			// -X,-Y,-Z
			if (vit.idx(dim)-L < 0 && dot(N,V) > 0)
			{
				UT_Vector3i idx;
				dst->posToIndex(P - multiplier*timestep*N*dot(N,V), idx[0], idx[1], idx[2]);
				if (idx[dim] < 0) vit.setValue(val);
			}

			// +X,+Y,+Z
			if (vit.idx(dim)+L > res[dim] && dot(N,V) < 0)
			{
				UT_Vector3i idx;
				dst->posToIndex(P - multiplier*timestep*N*dot(N,V), idx[0], idx[1], idx[2]);
				if (idx[dim] > res[dim]) vit.setValue(val);
			}
		}
#endif
#if 0
		// inside boundary layer
		if (vit.x()-L < 0 || vit.x()+L > res[0]
		|| vit.y()-L < 0 || vit.y()+L > res[1]
		|| vit.z()-L < 0 || vit.z()+L > res[2])
		{
			UT_Vector3 P, Pold;
			dst->indexToPos(vit.x(), vit.y(), vit.z(), P);
			Pold = P - multiplier*timestep*vel->getValue(P);

			UT_Vector3i idx;
			dst->posToIndex(Pold, idx[0], idx[1], idx[2]);

			// trace outside grid
			if (idx[0] < 0 || idx[0] > res[0]
			|| idx[1] < 0 || idx[1] > res[1]
			|| idx[2] < 0 || idx[2] > res[2])
			{
				vit.setValue(src->getValue(P));
			}
		}
#endif
	}
}
