/***
* MIT License
*
* Copyright (c) 2016 Morten Bojsen-Hansen
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
***/

#include "SIM_PMLAdvect.h"

#include <PRM/PRM_Include.h>
#include <SIM/SIM_DopDescription.h>
#include <SIM/SIM_ScalarField.h>
#include <SIM/SIM_VectorField.h>
#include <UT/UT_VectorTypes.h>
#include <UT/UT_Console.h>
#include <UT/UT_Performance.h>

// BaseClass created by DECLARE_DATAFACTORY
SIM_PMLAdvect::SIM_PMLAdvect(const SIM_DataFactory* factory)
	: BaseClass(factory)
{
}

SIM_PMLAdvect::~SIM_PMLAdvect()
{
}

// set up DOP (including paramaters etc.)
const SIM_DopDescription* SIM_PMLAdvect::getDopDescription()
{
	static PRM_Name velocityFieldName(GAS_NAME_VELOCITY, "Velocity Field");
	static PRM_Name splitVelocityFieldsName(PML_SPLIT_VELOCITY, "Split Velocity Fields");
	static PRM_Name splitDensityFieldsName(PML_SPLIT_DENSITY, "Split Density Fields");
	static PRM_Name layerThicknessName(PML_LAYER_THICKNESS, "Layer Thickness");
	static PRM_Name maximumDampingName(PML_MAXIMUM_DAMPING, "Maximum Damping");
	static PRM_Name dampingExponentName(PML_DAMPING_EXPONENT, "Damping Exponent");
	static PRM_Name fancyIntegratorName(PML_FANCY_INTEGRATOR, "Fancy Integrator");
	static PRM_Name dampingName(PML_DAMPING, "Damping");
	static PRM_Name dampPlusXName(PML_DAMP_PLUS_X, "+X");
	static PRM_Name dampMinusXName(PML_DAMP_MINUS_X, "-X");
	static PRM_Name dampPlusYName(PML_DAMP_PLUS_Y, "+Y");
	static PRM_Name dampMinusYName(PML_DAMP_MINUS_Y, "-Y");
	static PRM_Name dampPlusZName(PML_DAMP_PLUS_Z, "+Z");
	static PRM_Name dampMinusZName(PML_DAMP_MINUS_Z, "-Z");
	static PRM_Name cflCondName("cflcond", "CFL Condition");
	static PRM_Name contractSourceFieldsName(PML_CONTRACT_FIELDSOURCE, "Contract Source Fields");

	static PRM_Default layerThicknessDefault(6);
	static PRM_Default maximumDampingDefault(1000.0);
	static PRM_Default dampingExponentDefault(3.0);
	static PRM_Default cflCondDefault(1.0);

	static PRM_Template theTemplates[] = {
		PRM_Template(PRM_STRING, 1, &velocityFieldName),
		PRM_Template(PRM_STRING, 1, &splitVelocityFieldsName),
		PRM_Template(PRM_STRING, 1, &splitDensityFieldsName),
		PRM_Template(PRM_INT, 1, &layerThicknessName, &layerThicknessDefault),
		PRM_Template(PRM_FLT, 1, &maximumDampingName, &maximumDampingDefault),
		PRM_Template(PRM_FLT, 1, &dampingExponentName, &dampingExponentDefault),
		PRM_Template(PRM_TOGGLE, 1, &fancyIntegratorName, PRMoneDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampingName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampPlusXName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampMinusXName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampPlusYName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampMinusYName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampPlusZName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE, 1, &dampMinusZName, PRMzeroDefaults),
		PRM_Template(PRM_FLT, 1, &cflCondName, &cflCondDefault),
		PRM_Template(PRM_STRING, 1, &contractSourceFieldsName),
		PRM_Template()
	};

	static SIM_DopDescription theDopDescription(
		true, // make a DOP
		"sim_pmladvect", // internal name
		"PML Advect", // DOP name
		"$OS", // Default data name
		classname(), // the type of this DOP, usually the class.
		theTemplates); // template list for generating the DOP

	theDopDescription.setPureApplyData(true); // no gray inputs

	return &theDopDescription;
}

class AutoTimedEvent
{
	int eventId;

public:
	AutoTimedEvent(const SIM_Solver *solver, const std::string &str)
	{
		eventId = UTgetPerformance()->startTimedSolveEvent(solver, str.c_str());
	}

	~AutoTimedEvent()
	{
		UTgetPerformance()->stopEvent(eventId);
	}
};

bool SIM_PMLAdvect::solveGasSubclass(SIM_Engine &engine, SIM_Object *obj, SIM_Time time, SIM_Time timestep)
{
	UT_Interrupt *boss = UTgetInterrupt();
	SIM_VectorField *velocity = getVectorField(obj, GAS_NAME_VELOCITY);

	if (!velocity)
	{
		std::cout << GAS_NAME_VELOCITY << " not found" << std::endl;
		return false;
	}

	// compute timestep
	auto max = [](double x, double y, double z) { return std::max(x, std::max(y,z)); };
	int numSubSteps;
	{
		AutoTimedEvent(this, "Compute substeps");

		double velocityMaxAbs = max(
				velocity->getField(0)->reduceOp(SIM_RawField::REDUCE_MAXABS),
				velocity->getField(1)->reduceOp(SIM_RawField::REDUCE_MAXABS),
				velocity->getField(2)->reduceOp(SIM_RawField::REDUCE_MAXABS));
		double voxelSize = max(
			velocity->getVoxelSize(0).maxComponent(),
			velocity->getVoxelSize(1).maxComponent(),
			velocity->getVoxelSize(2).maxComponent());
		double maxTimestep = getCFLCondition() * voxelSize / (velocityMaxAbs + 1e-6);
		numSubSteps = std::ceil(timestep/maxTimestep);
	}

	if (numSubSteps > 20)
	{
		std::cout << "SIM_PMLAdvect warning: Doing " << numSubSteps << " substeps!" << std::endl;
	}

	SIM_VectorField *src;
	{
		AutoTimedEvent(this, "Duplicate source field");

		src = SIM_DATA_CREATE(*this, "src", SIM_VectorField, SIM_DATA_RETURN_EXISTING);
		src->matchField(velocity, true);
		for (int dim : {0,1,2}) *src->getField(dim) = *velocity->getField(dim);
	}

	SIM_DataArray splitVelocityFields, splitDensityFields;
	getMatchingData(splitVelocityFields, obj, PML_SPLIT_VELOCITY);
	getMatchingData(splitDensityFields, obj, PML_SPLIT_DENSITY);

	// class SIM_PMLContract;
	// SIM_PMLContract *tmp = SIM_DATA_CREATE(*this, "contract", SIM_PMLContract, 0);

	for (int step=0; step<numSubSteps; ++step)
	{
		if (boss->opInterrupt()) break;

		if (splitVelocityFields.entries() == 3)
		{
			AutoTimedEvent(this, "Advect velocity fields");

			for (int q : {0,1,2}) // beta
			{

				SIM_VectorField* dst = SIM_DATA_CAST(splitVelocityFields(q), SIM_VectorField);

				// nothing to do
				if (dst == nullptr) continue;

				for (int dim : {0,1,2}) // lambda
				{
					advectVelocityField(
						dst->getField(dim),
						src->getField(dim),
						velocity,
						dim,
						q,
						timestep/numSubSteps);
				}

				// flag dirty
				dst->pubHandleModification();
			}

			// it would really be better to reuse SIM_PMLContract here...
			// copy+pasta from SIM_PMLContract
			if (bool do_contraction = true)
			{
				AutoTimedEvent(this, "Contract");

				SIM_DataArray input;
				getMatchingData(input, obj, PML_CONTRACT_FIELDSOURCE);

				for (int dim : {0,1,2})
				{
					std::vector<const SIM_RawField*> sources;
					for (int i=0, I=input.entries(); i<I; ++i)
					{
						SIM_VectorField *src = SIM_DATA_CAST(input(i), SIM_VectorField);
						if (src) sources.push_back(src->getField(dim));
					}

					contract(src->getField(dim), sources);
					// enforceBoundary(src->getField(dim), dim);
					// src->getField(dim)->enforceBoundary();
				}
			}
		}

		if (splitDensityFields.entries() == 3)
		{
			AutoTimedEvent(this, "Advect density fields");

			for (int q : {0,1,2}) // beta
			{
				SIM_ScalarField* dst = SIM_DATA_CAST(splitDensityFields(q), SIM_ScalarField);

				// nothing to do
				if (dst == nullptr) continue;

				advectDensityField(
					dst->getField(),
					src->getField(q),
					q,
					timestep/numSubSteps);

				// flag dirty
				dst->pubHandleModification();
			}
		}
	}

	removeNamedSubData("src");

	return true;
}

double SIM_PMLAdvect::sigma(double x)
{
	if (x <= 0) return 0;

	double exp = getDampingExponent();
	double sigmamax = getMaximumDamping();

	// exp(1-1/x)
	// return sigmamax*std::exp(1 - 1/(x+1e-6));

	// return sigmamax*0.5*(std::tanh(x) + 1);

	// seems to work better than the two above
	// x^exp
	return sigmamax*std::pow(x, exp);
}

void SIM_PMLAdvect::advectVelocityFieldPartial(SIM_RawField *dst, const SIM_RawField *src, const SIM_VectorField *vel, unsigned dim, unsigned q, SIM_Time timestep, const UT_JobInfo &info)
{
	UT_Interrupt *boss = UTgetInterrupt();

	UT_VoxelArrayIteratorF vit;
	vit.setArray(dst->fieldNC());
	vit.setPartialRange(info.job(), info.numJobs());

	UT_VoxelProbeCubeF srcProbe;
	UT_Vector3 voxelSize(src->getVoxelSize());
	UT_Vector3T<bool> dampPlus(getDampPlusX(), getDampPlusY(), getDampPlusZ());
	UT_Vector3T<bool> dampMinus(getDampMinusX(), getDampMinusY(), getDampMinusZ());

	srcProbe.setPlusArray(src->field());

	double L = getLayerThickness();
	double N;
	double offset;

	if (q==dim) // staggered
	{
		N = vit.getArray()->getRes(q) - 1;
		offset = 0;
	}
	else // colocated
	{
		N = vit.getArray()->getRes(q);
		offset = 0.5;
	}

	for (vit.rewind(); !vit.atEnd(); vit.advance())
	{
		if (vit.isStartOfTile() && boss->opInterrupt()) break;

		srcProbe.setIndexPlus(vit);

		double sigmav=0, minus=0, plus=0;
		if (getDamping() && dampMinus[q]) minus = sigma( (L - (vit.idx(q) + offset)) / L );
		if (getDamping() && dampPlus[q]) plus = sigma( ((vit.idx(q) + offset) - (N-L)) / L );
		sigmav = std::max(minus, plus);

		double val = vit.getValue();
		UT_Vector3 P;
		src->indexToPos(vit.x(), vit.y(), vit.z(), P);
		UT_Vector3 V(vel->getValue(P));

		// biased gradient
		double upwind = V[q] > 0
			? srcProbe.getValue(0,0,0) - srcProbe.getValue(q==0?-1:0, q==1?-1:0, q==2?-1:0)
			: srcProbe.getValue(q==0?+1:0, q==1?+1:0, q==2?+1:0) - srcProbe.getValue(0,0,0);
		upwind /= voxelSize[q];

		// apply damping
		if (sigmav > 0)
		{
			if (getFancyIntegrator())
			{
				// PML (eq. 43+46)
				val -= timestep*sigmav*std::exp(-sigmav*timestep)*val;

				// PML (eq. 45)
				double a = V[q]*std::exp(-sigmav*timestep);
				val -= timestep*a*upwind;
			}
			else
			{
				val -= timestep*V[q]*upwind;
				val *= std::exp(-sigmav*timestep);
			}
		}
		else
		{
			val -= timestep*V[q]*upwind;
		}

		// finally update q
		vit.setValue(val);
	}
}

void SIM_PMLAdvect::advectDensityFieldPartial(SIM_RawField *dst, const SIM_RawField *src, unsigned q, SIM_Time timestep, const UT_JobInfo &info)
{
	UT_Interrupt *boss = UTgetInterrupt();

	UT_VoxelArrayIteratorF vit;
	vit.setArray(dst->fieldNC());
	vit.setPartialRange(info.job(), info.numJobs());

	UT_VoxelProbeCubeF srcProbe;
	UT_Vector3 voxelSize(src->getVoxelSize());
	UT_Vector3T<bool> dampPlus(getDampPlusX(), getDampPlusY(), getDampPlusZ());
	UT_Vector3T<bool> dampMinus(getDampMinusX(), getDampMinusY(), getDampMinusZ());

	srcProbe.setPlusArray(src->field());

	double L = getLayerThickness();
	double N = vit.getArray()->getRes(q);
	constexpr double offset = 0.5;

	for (vit.rewind(); !vit.atEnd(); vit.advance())
	{
		if (vit.isStartOfTile() && boss->opInterrupt()) break;

		srcProbe.setIndexPlus(vit);

		double val = vit.getValue();
		double dx = srcProbe.getValue(q==0?+1:0, q==1?+1:0, q==2?+1:0) - srcProbe.getValue(0,0,0);
		dx /= voxelSize[q];

		double sigmav=0, minus=0, plus=0;
		if (getDamping() && dampMinus[q]) minus = sigma( (L - (vit.idx(q) + offset)) / L );
		if (getDamping() && dampPlus[q]) plus = sigma( ((vit.idx(q) + offset) - (N-L)) / L );
		sigmav = std::max(minus, plus);

		if (sigmav > 0)
		{
			if (getFancyIntegrator())
			{
				// PML (eq. 43+46)
				val -= timestep*sigmav*std::exp(-sigmav*timestep)*val;

				// PML (eq. 45)
				double a = std::exp(-sigmav*timestep);
				val -= timestep*a*dx;
			}
			else
			{
				val -= timestep*dx;
				val *= std::exp(-sigmav*timestep);
			}
		}
		else
		{
			val -= timestep*dx;
		}

		// finally update q
		vit.setValue(val);
	}
}

// copy+pasta from SIM_PMLContract
void SIM_PMLAdvect::contractPartial(SIM_RawField *dst, const std::vector<const SIM_RawField*> &src, const UT_JobInfo &info)
#if 0
{
	UT_Interrupt *boss = UTgetInterrupt();

	UT_VoxelArrayIteratorF vit;
	vit.setArray(dst->fieldNC());
	vit.setPartialRange(info.job(), info.numJobs());

	std::vector<UT_VoxelProbeF> probes(src.size());
	for (int i=0, I=src.size(); i<I; ++i) probes[i].setConstArray(src[i]->field());

	for (vit.rewind(); !vit.atEnd(); vit.advance())
	{
		if (vit.isStartOfTile() && boss->opInterrupt()) break;

		fpreal val = 0;

		for (int i=0, I=src.size(); i<I; ++i)
		{
			probes[i].setIndex(vit);
			val += probes[i].getValue();
		}

		vit.setValue(val);
	}
}
#else
{
	UT_Interrupt *boss = UTgetInterrupt();

	UT_VoxelArrayIteratorF vit;
	vit.setArray(dst->fieldNC());
	vit.setPartialRange(info.job(), info.numJobs());

	std::vector<UT_VoxelProbeF> probes(src.size());
	for (int i=0, I=src.size(); i<I; ++i)
	if (dst->isAligned(src[i]))
	{
		probes[i].setConstArray(src[i]->field());
	}

	for (vit.rewind(); !vit.atEnd(); vit.advance())
	{
		if (vit.isStartOfTile() && boss->opInterrupt()) break;

		fpreal val = 0;

		for (int i=0, I=src.size(); i<I; ++i)
		{
			if (probes[i].isValid())
			{
				probes[i].setIndex(vit);
				val += probes[i].getValue();
			}
			else
			{
				UT_Vector3 P;
				dst->indexToPos(vit.x(), vit.y(), vit.z(), P);
				val += src[i]->getValue(P);
			}
		}

		vit.setValue(val);
	}
}
#endif

// copy+pasta from SIM_PMLContract
void SIM_PMLAdvect::enforceBoundaryPartial(SIM_RawField *dst, int dim, const UT_JobInfo &info)
{
	UT_Interrupt *boss = UTgetInterrupt();

	UT_VoxelArrayIteratorF vit;
	vit.setArray(dst->fieldNC());
	vit.setPartialRange(info.job(), info.numJobs());

	UT_Vector3T<bool> dampPlus(getDampPlusX(), getDampPlusY(), getDampPlusZ());
	UT_Vector3T<bool> dampMinus(getDampMinusX(), getDampMinusY(), getDampMinusZ());

	int N = vit.getArray()->getRes(dim);
	for (vit.rewind(); !vit.atEnd(); vit.advance())
	{
		if (vit.isStartOfTile() && boss->opInterrupt()) break;

		// if damping is NOT enabled, make sure to zero out boundary (so there is no inflow)
		if ((vit.idx(dim) == 0 && !dampMinus[dim]) || (vit.idx(dim) == N-1 && !dampPlus[dim]))
		{
			vit.setValue(0);
		}
		else if (!getDamping())
		{
			vit.setValue(0);
		}
	}
}
