/***
* MIT License
*
* Copyright (c) 2016 Morten Bojsen-Hansen
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
***/

#include "SIM_PMLComputeDivergence.h"

#include <PRM/PRM_Include.h>
#include <SIM/SIM_DopDescription.h>
#include <SIM/SIM_ScalarField.h>

// BaseClass created by DECLARE_DATAFACTORY
SIM_PMLComputeDivergence::SIM_PMLComputeDivergence(const SIM_DataFactory* factory)
	: BaseClass(factory)
{
}

SIM_PMLComputeDivergence::~SIM_PMLComputeDivergence()
{
}

// set up DOP (including paramaters etc.)
const SIM_DopDescription* SIM_PMLComputeDivergence::getDopDescription()
{
	static PRM_Name divergenceFieldName(GAS_NAME_DIVERGENCE, "Divergence Field");
	static PRM_Name splitDensityFieldsName(PML_SPLIT_DENSITY, "Split Density Fields");
	static PRM_Name layerThicknessName(PML_LAYER_THICKNESS, "Layer Thickness");
	static PRM_Name maximumDampingName(PML_MAXIMUM_DAMPING, "Maximum Damping");
	static PRM_Name dampingExponentName(PML_DAMPING_EXPONENT, "Damping Exponent");
	static PRM_Name dampingName(PML_DAMPING, "Damping");
	static PRM_Name dampPlusXName(PML_DAMP_PLUS_X, "+X");
	static PRM_Name dampMinusXName(PML_DAMP_MINUS_X, "-X");
	static PRM_Name dampPlusYName(PML_DAMP_PLUS_Y, "+Y");
	static PRM_Name dampMinusYName(PML_DAMP_MINUS_Y, "-Y");
	static PRM_Name dampPlusZName(PML_DAMP_PLUS_Z, "+Z");
	static PRM_Name dampMinusZName(PML_DAMP_MINUS_Z, "-Z");

	static PRM_Default layerThicknessDefault(6);
	static PRM_Default maximumDampingDefault(1000.0);
	static PRM_Default dampingExponentDefault(3.0);
	static PRM_Template theTemplates[] = {

		PRM_Template(PRM_STRING, 1, &divergenceFieldName),
		PRM_Template(PRM_STRING, 1, &splitDensityFieldsName),
		PRM_Template(PRM_INT, 1, &layerThicknessName, &layerThicknessDefault),
		PRM_Template(PRM_FLT, 1, &maximumDampingName, &maximumDampingDefault),
		PRM_Template(PRM_FLT, 1, &dampingExponentName, &dampingExponentDefault),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampingName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampPlusXName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampMinusXName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampPlusYName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampMinusYName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampPlusZName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE, 1, &dampMinusZName, PRMzeroDefaults),
		PRM_Template()
	};

	static SIM_DopDescription theDopDescription(
		true, // make a DOP
		"sim_pmlcomputedivergence", // internal name
		"PML Compute Divergence", // DOP name
		"$OS", // Default data name
		classname(), // the type of this DOP, usually the class.
		theTemplates); // template list for generating the DOP

	theDopDescription.setPureApplyData(true); // no gray inputs

	return &theDopDescription;
}

bool SIM_PMLComputeDivergence::solveGasSubclass(SIM_Engine &engine, SIM_Object *obj, SIM_Time time, SIM_Time timestep)
{
	UT_Interrupt *boss = UTgetInterrupt();

	SIM_ScalarField *dst = getScalarField(obj, GAS_NAME_DIVERGENCE);
	if (!dst) return false;

	SIM_DataArray splitDensityFields;
	getMatchingData(splitDensityFields, obj, PML_SPLIT_DENSITY);
	if (splitDensityFields.entries() != 3) return false;

	const SIM_RawField* src[3];
	for (int q : {0,1,2})
	{
		SIM_ScalarField *tmp = SIM_DATA_CAST(splitDensityFields(q), SIM_ScalarField);
		if (!tmp) return false;
		src[q] = tmp->getField();
	}

	contract(dst->getField(), src);

	dst->pubHandleModification();

	return true;
}

double SIM_PMLComputeDivergence::sigma(double x)
{
	if (x <= 0) return 0;

	double exp = getDampingExponent();
	double sigmamax = getMaximumDamping();

	// exp(1-1/x)
	// return sigmamax*std::exp(1 - 1/(x+1e-6));

	// return sigmamax*0.5*(std::tanh(x) + 1);

	// seems to work better than the two above
	// x^exp
	return sigmamax*std::pow(x, exp);
}

void SIM_PMLComputeDivergence::contractPartial(SIM_RawField *dst, const SIM_RawField* src[3], const UT_JobInfo &info)
{
	UT_Interrupt *boss = UTgetInterrupt();

	UT_VoxelArrayIteratorF vit;
	vit.setArray(dst->fieldNC());
	vit.setPartialRange(info.job(), info.numJobs());

	UT_Vector3T<bool> dampPlus(getDampPlusX(), getDampPlusY(), getDampPlusZ());
	UT_Vector3T<bool> dampMinus(getDampMinusX(), getDampMinusY(), getDampMinusZ());
	UT_VoxelProbeF probes[3];
	for (int q : {0,1,2}) probes[q].setConstArray(src[q]->field());

	double L = getLayerThickness();
	constexpr double offset = 0.5;

	for (vit.rewind(); !vit.atEnd(); vit.advance())
	{
		if (vit.isStartOfTile() && boss->opInterrupt()) break;

		fpreal val = 0;

		for (int q : {0,1,2})
		{
			double N = vit.getArray()->getRes(q);
			probes[q].setIndex(vit);

			double sigmav=0, minus=0, plus=0;
			if (getDamping() && dampMinus[q]) minus = sigma( (L - (vit.idx(q) + offset)) / L );
			if (getDamping() && dampPlus[q]) plus = sigma( ((vit.idx(q) + offset) - (N-L)) / L );
			sigmav = std::max(minus, plus);
			val -= sigmav*probes[q].getValue();
		}

		vit.setValue(val);
	}
}
