/***
* MIT License
*
* Copyright (c) 2016 Morten Bojsen-Hansen
*
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
*
* The above copyright notice and this permission notice shall be included in all
* copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
* SOFTWARE.
***/

#include "SIM_PMLContract.h"

#include <PRM/PRM_Include.h>
#include <SIM/SIM_DopDescription.h>
#include <SIM/SIM_RawField.h>
#include <SIM/SIM_ScalarField.h>
#include <SIM/SIM_VectorField.h>
#include <SIM/SIM_MatrixField.h>

// BaseClass created by DECLARE_DATAFACTORY
SIM_PMLContract::SIM_PMLContract(const SIM_DataFactory* factory)
	: BaseClass(factory)
{
}

SIM_PMLContract::~SIM_PMLContract()
{
}

// set up DOP (including paramaters etc.)
const SIM_DopDescription* SIM_PMLContract::getDopDescription()
{
	static PRM_Name destinationFieldName(GAS_NAME_FIELDDEST, "Destination Field");
	static PRM_Name sourceFieldsName(GAS_NAME_FIELDSOURCE, "Source Fields");
	static PRM_Name dampingName(PML_DAMPING, "Damping");
	static PRM_Name dampPlusXName(PML_DAMP_PLUS_X, "+X");
	static PRM_Name dampMinusXName(PML_DAMP_MINUS_X, "-X");
	static PRM_Name dampPlusYName(PML_DAMP_PLUS_Y, "+Y");
	static PRM_Name dampMinusYName(PML_DAMP_MINUS_Y, "-Y");
	static PRM_Name dampPlusZName(PML_DAMP_PLUS_Z, "+Z");
	static PRM_Name dampMinusZName(PML_DAMP_MINUS_Z, "-Z");


	static PRM_Template theTemplates[] = {
		PRM_Template(PRM_STRING, 1, &destinationFieldName),
		PRM_Template(PRM_STRING, 1, &sourceFieldsName),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampingName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampPlusXName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampMinusXName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampPlusYName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampMinusYName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE | PRM_TYPE_JOIN_NEXT, 1, &dampPlusZName, PRMzeroDefaults),
		PRM_Template(PRM_TOGGLE, 1, &dampMinusZName, PRMzeroDefaults),
		PRM_Template()
	};

	static SIM_DopDescription theDopDescription(
		true, // make a DOP
		"sim_pmlcontract", // internal name
		"PML Contract", // DOP name
		"$OS", // Default data name
		classname(), // the type of this DOP, usually the class.
		theTemplates); // template list for generating the DOP

	theDopDescription.setPureApplyData(true); // no gray inputs

	return &theDopDescription;
}

bool SIM_PMLContract::solveGasSubclass(SIM_Engine &engine, SIM_Object *obj, SIM_Time time, SIM_Time timestep)
{
	if (SIM_ScalarField *dst = getScalarField(obj, GAS_NAME_FIELDDEST))
	{
		SIM_DataArray input;
		getMatchingData(input, obj, GAS_NAME_FIELDSOURCE);

		std::vector<const SIM_RawField*> sources;
		for (int i=0, I=input.entries(); i<I; ++i)
		{
			SIM_ScalarField *src = SIM_DATA_CAST(input(i), SIM_ScalarField);
			if (src) sources.push_back(src->getField());
		}

		contract(dst->getField(), sources);
		dst->pubHandleModification();
	}

	else if (SIM_VectorField *dst = getVectorField(obj, GAS_NAME_FIELDDEST))
	{
		SIM_DataArray input;
		getMatchingData(input, obj, GAS_NAME_FIELDSOURCE);

		for (int dim : {0,1,2})
		{
			std::vector<const SIM_RawField*> sources;
			for (int i=0, I=input.entries(); i<I; ++i)
			{
				SIM_VectorField *src = SIM_DATA_CAST(input(i), SIM_VectorField);
				if (src) sources.push_back(src->getField(dim));
			}

			contract(dst->getField(dim), sources);
			// enforceBoundary(dst->getField(dim), dim);
			// dst->getField(dim)->enforceBoundary();
		}

		dst->pubHandleModification();
	}

	else if (SIM_MatrixField *dst = getMatrixField(obj, GAS_NAME_FIELDDEST))
	{
		return false; // TODO
	}

	return true;
}

void SIM_PMLContract::contractPartial(SIM_RawField *dst, const std::vector<const SIM_RawField*> &src, const UT_JobInfo &info)
#if 0
{
	UT_Interrupt *boss = UTgetInterrupt();

	UT_VoxelArrayIteratorF vit;
	vit.setArray(dst->fieldNC());
	vit.setPartialRange(info.job(), info.numJobs());

	std::vector<UT_VoxelProbeF> probes(src.size());
	for (int i=0, I=src.size(); i<I; ++i) probes[i].setConstArray(src[i]->field());

	for (vit.rewind(); !vit.atEnd(); vit.advance())
	{
		if (vit.isStartOfTile() && boss->opInterrupt()) break;

		fpreal val = 0;

		for (int i=0, I=src.size(); i<I; ++i)
		{
			probes[i].setIndex(vit);
			val += probes[i].getValue();
		}

		vit.setValue(val);
	}
}
#else
{
	UT_Interrupt *boss = UTgetInterrupt();

	UT_VoxelArrayIteratorF vit;
	vit.setArray(dst->fieldNC());
	vit.setPartialRange(info.job(), info.numJobs());

	std::vector<UT_VoxelProbeF> probes(src.size());
	for (int i=0, I=src.size(); i<I; ++i)
	if (dst->isAligned(src[i]))
	{
		probes[i].setConstArray(src[i]->field());
	}

	for (vit.rewind(); !vit.atEnd(); vit.advance())
	{
		if (vit.isStartOfTile() && boss->opInterrupt()) break;

		fpreal val = 0;

		for (int i=0, I=src.size(); i<I; ++i)
		{
			if (probes[i].isValid())
			{
				probes[i].setIndex(vit);
				val += probes[i].getValue();
			}
			else
			{
				UT_Vector3 P;
				dst->indexToPos(vit.x(), vit.y(), vit.z(), P);
				val += src[i]->getValue(P);
			}
		}

		vit.setValue(val);
	}
}
#endif

void SIM_PMLContract::enforceBoundaryPartial(SIM_RawField *dst, int dim, const UT_JobInfo &info)
{
	UT_Interrupt *boss = UTgetInterrupt();

	UT_VoxelArrayIteratorF vit;
	vit.setArray(dst->fieldNC());
	vit.setPartialRange(info.job(), info.numJobs());

	UT_Vector3T<bool> dampPlus(getDampPlusX(), getDampPlusY(), getDampPlusZ());
	UT_Vector3T<bool> dampMinus(getDampMinusX(), getDampMinusY(), getDampMinusZ());

	int N = vit.getArray()->getRes(dim);
	for (vit.rewind(); !vit.atEnd(); vit.advance())
	{
		if (vit.isStartOfTile() && boss->opInterrupt()) break;

		// if damping is NOT enabled, make sure to zero out boundary (so there is no inflow)
		if ((vit.idx(dim) == 0 && (!getDamping() || !dampMinus[dim])) || (vit.idx(dim) == N-1 && (!getDamping() || !dampPlus[dim])))
		{
			vit.setValue(0);
		}
	}
}
